
C    **************************************************************************
C    *                                                                        *
C    *                  *****     SUBROUTINE FOURT     *****                  *
C    *                                                                        *
C    *    THE COOLEY-TUKEY FAST FOURIER TRANSFORM IN USASI BASIC FORTRAN      *
C    *                                                                        *
C    *    TRANSFORM(K1,K2,...) = SUM(DATA(J1,J2,...)*EXP(ISIGN*2*PI*SQRT(-1)  *
C    *     *((J1-1)*(K1-1)/NN(1)+(J2-1)*(K2-1)/NN(2)+...))), SUMMED FOR ALL   *
C    *     J1, K1 BETWEEN 1 AND NN(1), J2, K2 BETWEEN 1 AND NN(2), ETC.       *
C    *     THERE IS NO LIMIT TO THE NUMBER OF SUBSCRIPTS.  DATA IS A          *
C    *     MULTIDIMENSIONAL COMPLEX ARRAY WHOSE REAL AND IMAGINARY            *
C    *     PARTS ARE ADJACENT IN STORAGE, SUCH AS FORTRAN IV PLACES THEM.     *
C    *     IF ALL IMAGINARY PARTS ARE ZERO (DATA ARE DISGUISED REAL), SET     *
C    *     IFORM TO ZERO TO CUT THE RUNNING TIME BY UP TO FORTY PERCENT.      *
C    *     OTHERWISE, IFORM = +1.  THE LENGTHS OF ALL DIMENSIONS ARE          *
C    *     STORED IN ARRAY NN, OF LENGTH NDIM.  THEY MAY BE ANY POSITIVE      *
C    *     INTEGERS, THO THE PROGRAM RUNS FASTER ON COMPOSITE INTEGERS, AND   *
C    *     ESPECIALLY FAST ON NUMBERS RICH IN FACTORS OF TWO.  ISIGN IS +1    *
C    *     OR -1.  IF A -1 TRANSFORM IS FOLLOWED BY A +1 ONE (OR A +1         *
C    *     BY A -1) THE ORIGINAL DATA REAPPEAR, MULTIPLIED BY NTOT (=NN(1)*   *
C    *     NN(2)*...).  TRANSFORM VALUES ARE ALWAYS COMPLEX, AND ARE RETURNED *
C    *     IN ARRAY DATA, REPLACING THE INPUT.  IN ADDITION, IF ALL           *
C    *     DIMENSIONS ARE NOT POWERS OF TWO, ARRAY WORK MUST BE SUPPLIED,     *
C    *     COMPLEX OF LENGTH EQUAL TO THE LARGEST NON 2**K DIMENSION.         *
C    *     OTHERWISE, REPLACE WORK BY ZERO IN THE CALLING SEQUENCE.           *
C    *     NORMAL FORTRAN DATA ORDERING IS EXPECTED, FIRST SUBSCRIPT VARYING  *
C    *     FASTEST.  ALL SUBSCRIPTS BEGIN AT ONE.                             *
C    *                                                                        *
C    *    RUNNING TIME IS MUCH SHORTER THAN THE NAIVE NTOT**2, BEING          *
C    *     GIVEN BY THE FOLLOWING FORMULA.  DECOMPOSE NTOT INTO               *
C    *     2**K2 * 3**K3 * 5**K5 * ....  LET SUM2 = 2*K2, SUMF = 3*K3 + 5*K5  *
C    *     + ... AND NF = K3 + K5 + ....  THE TIME TAKEN BY A MULTI-          *
C    *     DIMENSIONAL TRANSFORM ON THESE NTOT DATA IS T = T0 + NTOT*(T1+     *
C    *     T2*SUM2+T3*SUMF+T4*NF).  ON THE CDC 3300 (FLOATING POINT ADD TIME  *
C    *     OF SIX MICROSECONDS), T = 3000 + NTOT*(500+43*SUM2+68*SUMF+        *
C    *     320*NF) MICROSECONDS ON COMPLEX DATA.  IN ADDITION, THE            *
C    *     ACCURACY IS GREATLY IMPROVED, AS THE RMS RELATIVE ERROR IS         *
C    *     BOUNDED BY 3*2**(-B)*SUM(FACTOR(J)**1.5), WHERE B IS THE NUMBER    *
C    *     OF BITS IN THE FLOATING POINT FRACTION AND FACTOR(J) ARE THE       *
C    *     PRIME FACTORS OF NTOT.                                             *
C    *                                                                        *
C    *    PROGRAM BY NORMAN BRENNER FROM THE BASIC PROGRAM BY CHARLES         *
C    *     RADER.  RALPH ALTER SUGGESTED THE IDEA FOR THE DIGIT REVERSAL.     *
C    *     MIT LINCOLN LABORATORY, AUGUST 1967.  THIS IS THE FASTEST AND MOST *
C    *     VERSATILE VERSION OF THE FFT KNOWN TO THE AUTHOR.  SHORTER PRO-    *
C    *     GRAMS FOUR1 AND FOUR2 RESTRICT DIMENSION LENGTHS TO POWERS OF TWO. *
C    *     SEE-- IEEE AUDIO TRANSACTIONS (JUNE 1967), SPECIAL ISSUE ON FFT.   *
C    *                                                                        *
C    *    THE DISCRETE FOURIER TRANSFORM PLACES THREE RESTRICTIONS UPON THE   *
C    *     DATA.                                                              *
C    *     1.  THE NUMBER OF INPUT DATA AND THE NUMBER OF TRANSFORM VALUES    *
C    *         MUST BE THE SAME.                                              *
C    *     2.  BOTH THE INPUT DATA AND THE TRANSFORM VALUES MUST REPRESENT    *
C    *         EQUISPACED POINTS IN THEIR RESPECTIVE DOMAINS OF TIME AND      *
C    *         FREQUENCY.  CALLING THESE SPACINGS DELTAT AND DELTAF, IT       *
C    *         MUST BE TRUE THAT DELTAF=2*PI/(NN(I)*DELTAT).  OF COURSE,      *
C    *         DELTAT NEED NOT BE THE SAME FOR EVERY DIMENSION.               *
C    *     3.  CONCEPTUALLY AT LEAST, THE INPUT DATA AND THE TRANSFORM        *
C    *         OUTPUT REPRESENT SINGLE CYCLES OF PERIODIC FUNCTIONS.          *
C    *                                                                        *
C    *     EXAMPLE 1.  THREE-DIMENSIONAL FORWARD FOURIER TRANSFORM OF A       *
C    *       COMPLEX ARRAY DIMENSIONED 32 BY 25 BY 13 IN FORTRAN IV.          *
C    *       DIMENSION DATA(32,25,13),WORK(50),NN(3)                          *
C    *       COMPLEX DATA                                                     *
C    *       DATA NN/32,25,13/                                                *
C    *       DO 1 I=1,32                                                      *
C    *         DO 1 J=1,25                                                    *
C    *           DO 1 K=1,13                                                  *
C    *    1        DATA(I,J,K)=COMPLEX VALUE                                  *
C    *       CALL FOURT(DATA,NN,3,-1,1,WORK)                                  *
C    *                                                                        *
C    *     EXAMPLE 2.  ONE-DIMENSIONAL FORWARD TRANSFORM OF A REAL ARRAY OF   *
C    *       LENGTH 64 IN FORTRAN II.                                         *
C    *       DIMENSION DATA(2,64)                                             *
C    *       DO 2 I=1,64                                                      *
C    *         DATA(1,I)=REAL PART                                            *
C    *    2    DATA(2,I)=0.                                                   *
C    *       CALL FOURT(DATA,64,1,-1,0,work)                                  *
C    **************************************************************************
C
      SUBROUTINE FOURT (DATA, NN, NDIM, KSIGN, IFORM, WORK)
C
      DIMENSION DATA(1), NN(1), IFACT(32), WORK(1)  
C  ----------------------------------------------------------------------------
C
      TWOPI = 6.283185307         
      IF(NDIM-1)920,1,1         
    1      NTOT = 2
      DO 2 IDIM = 1,NDIM
      IF(NN(IDIM))920,920,2
    2      NTOT = NTOT*NN(IDIM)
C  **************************************
C  *    MAIN LOOP FOR EACH DIMENSION    *
C  **************************************
      NP1 = 2
      DO 910 IDIM = 1,NDIM
      N   = NN(IDIM)
      NP2 = NP1*N
      IF(N-1)920,900,5
C  ******************
C  *    FACTOR N    *
C  ******************
    5      M     = N
      NTWO  = NP1
      IF    = 1
      IDIV  = 2
   10      IQUOT = M/IDIV
      IREM  = M-IDIV*IQUOT
      IF(IQUOT-IDIV)50,11,11
   11      IF(IREM)20,12,20
   12      NTWO = NTWO+NTWO
      M    = IQUOT
      GO TO 10
   20      IDIV  = 3
   30      IQUOT = M/IDIV
      IREM  = M-IDIV*IQUOT
      IF(IQUOT-IDIV)60,31,31
   31      IF(IREM)40,32,40
   32      IFACT(IF) = IDIV
      IF        = IF+1
      M         = IQUOT
      GO TO 30
   40      IDIV = IDIV+2
      GO TO 30
   50      IF(IREM)60,51,60
   51      NTWO = NTWO+NTWO
      GO TO 70
   60      IFACT(IF) = M
C  ************************************************************************
C  *    SEPARATE FOUR CASES--                                             *
C  *      1. COMPLEX TRANSFORM OR REAL TRANSFORM FOR THE 4TH, 5TH,ETC.    *
C  *         DIMENSIONS.                                                  *
C  *      2. REAL TRANSFORM FOR THE 2ND OR 3RD DIMENSION.  METHOD--       *
C  *         TRANSFORM HALF THE DATA, SUPPLYING THE OTHER HALF BY CON-    *
C  *         JUGATE SYMMETRY.                                             *
C  *      3. REAL TRANSFORM FOR THE 1ST DIMENSION, N ODD.  METHOD--       *
C  *         TRANSFORM HALF THE DATA AT EACH STAGE, SUPPLYING THE OTHER   *
C  *         HALF BY CONJUGATE SYMMETRY.                                  *
C  *      4. REAL TRANSFORM FOR THE 1ST DIMENSION, N EVEN.  METHOD--      *
C  *         TRANSFORM A COMPLEX ARRAY OF LENGTH N/2 WHOSE REAL PARTS     *
C  *         ARE THE EVEN NUMBERED REAL VALUES AND WHOSE IMAGINARY PARTS  *
C  *         ARE THE ODD NUMBERED REAL VALUES.  SEPARATE AND SUPPLY       *
C  *         THE SECOND HALF BY CONJUGATE SYMMETRY.                       *
C  ************************************************************************
   70      NON2  = NP1*(NP2/NTWO)
      ICASE = 1
      IF(IDIM-4) 71,90,90
   71      IF(IFORM) 72,72,90
   72      ICASE = 2
      IF(IDIM-1) 73,73,90
   73      ICASE = 3
       IF(NTWO-NP1) 90,90,74
   74      ICASE = 4
      NTWO = NTWO/2
      N    = N/2
      NP2  = NP2/2
      NTOT = NTOT/2
      I    = 3
      DO J = 2,NTOT
        DATA(J) = DATA(I)
        I       = I+2
      ENDDO
   90      I1RNG = NP1
      IF(ICASE-2) 100,95,100
   95      I1RNG = NP0*(1+NPREV/2)
C  ***********************************************************************
C  *    SHUFFLE ON THE FACTORS OF TWO IN N.  AS THE SHUFFLING            *
C  *    CAN BE DONE BY SIMPLE INTERCHANGE, NO WORKING ARRAY IS NEEDED    *
C  ***********************************************************************
  100      IF(NTWO-NP1) 600,600,110
  110      NP2HF = NP2/2
      J     = 1
      DO 150 I2 = 1,NP2,NON2 
        IF(J-I2) 120,130,130
  120        I1MAX = I2+NON2-2
        DO I1 = I2,I1MAX,2
          DO I3 = I1,NTOT,NP2
            J3         = J+I3-I2
            TEMPR      = DATA(I3)
            TEMPI      = DATA(I3+1)
            DATA(I3)   = DATA(J3)
            DATA(I3+1) = DATA(J3+1)
            DATA(J3)   = TEMPR
            DATA(J3+1) = TEMPI
          ENDDO
        ENDDO
  130        M = NP2HF
  140        IF(J-M) 150,150,145
  145        J = J-M
        M = M/2
        IF(M-NON2) 150,140,140
  150      J = J+M
C  ****************************************************************************
C  *    MAIN LOOP FOR FACTORS OF TWO.  PERFORM FOURIER TRANSFORMS OF          *
C  *     LENGTH FOUR, WITH ONE OF LENGTH TWO IF NEEDED.  THE TWIDDLE FACTOR   *
C  *     W=EXP(ISIGN*2*PI*SQRT(-1)*M/(4*MMAX)).  CHECK FOR W=ISIGN*SQRT(-1)   *
C  *     AND REPEAT FOR W=ISIGN*SQRT(-1)*CONJUGATE(W).                        *
C  ****************************************************************************
      NON2T = NON2+NON2
      IPAR  = NTWO/NP1
  310      IF(IPAR-2) 350,330,320
  320      IPAR = IPAR/4
      GO TO 310
  330      DO I1 = 1,I1RNG,2
        DO J3 = I1,NON2,NP1
          DO K1 = J3,NTOT,NON2T
            K2         = K1+NON2
            TEMPR      = DATA(K2)
            TEMPI      = DATA(K2+1)
            DATA(K2)   = DATA(K1)-TEMPR
            DATA(K2+1) = DATA(K1+1)-TEMPI
            DATA(K1)   = DATA(K1)+TEMPR
            DATA(K1+1) = DATA(K1+1)+TEMPI
          ENDDO
        ENDDO
      ENDDO
  350      MMAX = NON2
  360      IF(MMAX-NP2HF) 370,600,600
  370      LMAX = MAX0(NON2T,MMAX/2)
      IF(MMAX-NON2) 405,405,380
  380      THETA = -TWOPI*FLOAT(NON2)/FLOAT(4*MMAX)
      IF (KSIGN) 400,390,390
  390      THETA = -THETA
  400      WR    = COS(THETA)
      WI    = SIN(THETA)
      WSTPR = -2.*WI*WI
      WSTPI =  2.*WR*WI
  405      DO 570 L = NON2,LMAX,NON2T
      M = L
      IF(MMAX-NON2) 420,420,410
  410      W2R = WR*WR-WI*WI
      W2I = 2.*WR*WI
      W3R = W2R*WR-W2I*WI
      W3I = W2R*WI+W2I*WR
  420      DO 530 I1 = 1,I1RNG,2
      DO 530 J3 = I1,NON2,NP1
      KMIN = J3+IPAR*M
      IF(MMAX-NON2) 430,430,440
  430      KMIN  = J3
  440      KDIF  = IPAR*MMAX
  450      KSTEP = 4*KDIF
      DO 520 K1 = KMIN,NTOT,KSTEP
      K2 = K1+KDIF
      K3 = K2+KDIF
      K4 = K3+KDIF
      IF(MMAX-NON2) 460,460,480
  460      U1R = DATA(K1)+DATA(K2)
      U1I = DATA(K1+1)+DATA(K2+1)
      U2R = DATA(K3)+DATA(K4)
      U2I = DATA(K3+1)+DATA(K4+1)
      U3R = DATA(K1)-DATA(K2)
      U3I = DATA(K1+1)-DATA(K2+1)
      IF (KSIGN) 470,475,475
  470      U4R = DATA(K3+1)-DATA(K4+1)
      U4I = DATA(K4)-DATA(K3)
      GO TO 510
  475      U4R = DATA(K4+1)-DATA(K3+1)
      U4I = DATA(K3)-DATA(K4)
      GO TO 510
  480      T2R = W2R*DATA(K2)-W2I*DATA(K2+1)
      T2I = W2R*DATA(K2+1)+W2I*DATA(K2)
      T3R = WR*DATA(K3)-WI*DATA(K3+1)
      T3I = WR*DATA(K3+1)+WI*DATA(K3)
      T4R = W3R*DATA(K4)-W3I*DATA(K4+1)
      T4I = W3R*DATA(K4+1)+W3I*DATA(K4)
      U1R = DATA(K1)+T2R
      U1I = DATA(K1+1)+T2I
      U2R = T3R+T4R
      U2I = T3I+T4I
      U3R = DATA(K1)-T2R
      U3I = DATA(K1+1)-T2I
      IF (KSIGN) 490,500,500
  490      U4R = T3I-T4I
      U4I = T4R-T3R
      GO TO 510
  500      U4R = T4I-T3I
      U4I = T3R-T4R
  510      DATA(K1)   = U1R+U2R
      DATA(K1+1) = U1I+U2I
      DATA(K2)   = U3R+U4R
      DATA(K2+1) = U3I+U4I
      DATA(K3)   = U1R-U2R
      DATA(K3+1) = U1I-U2I
      DATA(K4)   = U3R-U4R
  520      DATA(K4+1) = U3I-U4I
      KMIN       = 4*(KMIN-J3)+J3
      KDIF       = KSTEP
      IF(KDIF-NP2) 450,530,530
  530      CONTINUE
      M = MMAX-M
      IF (KSIGN) 540,550,550
  540      TEMPR = WR
      WR    = -WI
      WI    = -TEMPR
      GO TO 560
  550      TEMPR = WR
      WR    = WI
      WI    = TEMPR
  560      IF(M-LMAX) 565,565,410
  565      TEMPR = WR
      WR    = WR*WSTPR-WI*WSTPI+WR
  570      WI    = WI*WSTPR+TEMPR*WSTPI+WI
      IPAR  = 3-IPAR
      MMAX  = MMAX+MMAX
      GO TO 360
  600      CALL L6TO9(DATA,NN,NDIM,KSIGN,IFORM,WORK,NTWO,NPT2,
     1  NP2,NP0,KSTEP,I1MAX,IFACT,
     1  NON2,IF,NP1,ICASE,THETA,TWOPI,WSTPR,WSTPI,WR,WI,I1RNG,
     2  I1,I3,NTOT,J3,TEMPR,N,KMIN,TEMPI,I2,J,I,IDIM)
C  ***************************************
C  *    END OF LOOP ON EACH DIMENSION    *
C  ***************************************
  900      NP0   = NP1
      NP1   = NP2
  910      NPREV = N
  920      RETURN
      END
